#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This plugin handles the provisioning of Broadlink devices
# Taken from https://github.com/mjg59/python-broadlink
#
# Copyright (c) 2018 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
#
#
#
import asyncio
from struct import pack
import logging, socket


class Broadlink(object):

    name = "broadlink"

    def __init__(self, mac):
        """
        The go_on attribute must exist. It is set to False when provisioning is done

        :param mac: MAC address of the device being provisioned
        :type mac: str
        """
        self.go_on = True
        self.mac = mac

    @classmethod
    def can_handle(self, cells):
        """ Given a list of cell names, return a list of those it can handle

        :param cells: A list of cell names
        :type cmd: list
        :returns: A dictionary of those it can handle. The key is the cell's name, the value a dictionary
                  with a WIFI key "passwd" (empty is not needed) and IP addresses "ip" and "ipv6" (empty if use dhcp)
                  IP address must be of the form <address>/<bits>
        :rtype: dict
        """
        resu = {}
        for x in cells:
            if x.lower().strip().startswith("broadlink"):
                resu[x] = {"passwd": "", "ip": "", "ipv6": ""}

        return resu

    async def secure(self, user, passwd):
        """ Nothing here """
        await asyncio.sleep(0)

    async def set_options(self, options={}):
        """ Nothing here """
        await asyncio.sleep(0)

    def build_message(self, ssid, psk, ktype):
        """Build the provisioning payload

            :param ssid: The cell the device should connect to
            :type ssid: str
            :param psk: The key for the cell the device should connect to
            :type psk: str
            :param ktype: The key encyption type: wep, wpa, wpa2, none
            :type ktype: str
            :returns: the message as a bytearray
            :rtype: bytearray
        """
        payload = bytearray(0x88)
        payload[0x26] = 0x14  # This seems to always be set to 14
        # Add the SSID to the payload
        ssid_start = 68
        ssid_length = 0
        for letter in ssid:
            payload[(ssid_start + ssid_length)] = ord(letter)
            ssid_length += 1
        # Add the WiFi password to the payload
        pass_start = 100
        pass_length = 0
        for letter in psk:
            payload[(pass_start + pass_length)] = ord(letter)
            pass_length += 1
        #
        payload[0x84] = ssid_length  # Character length of SSID
        payload[0x85] = pass_length  # Character length of password
        payload[0x86] = ["none", "wep", "wpa", "wpa2"].index(
            ktype
        )  # Type of encryption (00=none,01=WEP,02=WPA1,03=WPA2,04=WPA1/2)
        #
        checksum = 0xBEAF
        for i in range(len(payload)):
            checksum += payload[i]
            checksum = checksum & 0xFFFF
        #
        payload[0x20] = checksum & 0xFF  # Checksum 1 position
        payload[0x21] = checksum >> 8  # Checksum 2 position

        return payload

    async def provision(self, ip, ssid, psk, ktype="none"):
        """Coroutine to perform provisioning

           As long as this method returns AGAIN, this coroutine must be called after switching
           wireless network if necessary. Each AGAIN shoud toggle between the device being provisioned
           cell and the original cell (nothig is done if the interface was not being used.

           If information is returnedits handling should be application specific.

            :param ip: IP address of the interface configured to access the device
            :type cmd: str
            :param ssid: The cell the device should connect to
            :type ssid: str
            :param psk: The key for the cell the device should connect to
            :type psk: str
            :param ktype: The key encyption type: wep, wpa, wpa2, none
            :type ktype: str
            :returns: a dictionary of information or AGAIN if needed
            :rtype: list
        """

        class BroadlinktProtocol:
            def __init__(self, payload, ip):
                self.payload = payload
                self.ip = ip

            def connection_made(self, transport):
                self.transport = transport
                sock = transport.get_extra_info("socket")
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                self.broadcast()

            def datagram_received(self, data, addr):
                logging.debug("data received: {}, {}".format(data, addr))

            def broadcast(self):
                self.transport.sendto(
                    payload, (".".join(self.ip.split(".")[:-1] + ["255"]), 80)
                )

        payload = self.build_message(ssid, psk, ktype)
        try:
            loop = asyncio.get_running_loop()
        except:
            loop = asyncio.get_event_loop()
        coro = loop.create_datagram_endpoint(
            lambda: BroadlinktProtocol(payload, ip), local_addr=("0.0.0.0", 8080)
        )
        xx = await coro
        await asyncio.sleep(3)
        self.go_on = False
        return {self.mac: {"type": "broadlink"}}


PluginObject = Broadlink



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This plugin handles the provisioning of TP-Link devices
#
# Copyright (c) 2018 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
#
#
#
import asyncio
from struct import pack
import logging
import json

# Encryption and Decryption of TP-Link Smart Home Protocol
# XOR Autokey Cipher with starting key = 171
def encrypt(string):
    key = 171
    result = pack(">I", len(string))
    for i in string:
        a = key ^ ord(i)
        key = a
        result += bytes([a])
    return result


def decrypt(string):
    key = 171
    result = ""
    for i in string:
        a = key ^ i
        key = i
        result += chr(a)
    return result


class TPLink(object):

    name = "tp-link"

    def __init__(self, mac):
        """
        The go_on attribute must exist. It is set to False when provisioning is done

        :param mac: MAC address of the device being provisioned
        :type mac: str

        """
        self.go_on = True
        self.mac = mac

    @classmethod
    def can_handle(self, cells):
        """ Given a list of cell names, return a list of those it can handle

        :param cells: A list of cell names
        :type cmd: list
        :returns: A dictionary of those it can handle. The key is the cell's name, the value a dictionary
                  with a WIFI key "passwd" (empty is not needed) and IP addresses "ip" and "ipv6" (empty if use dhcp)
                  IP address must be of the form <address>/<bits>
        :rtype: dict
        """
        resu = {}
        for x in cells:
            if x.lower().strip().startswith("tp-link"):
                resu[x] = {"passwd": "", "ip": "", "ipv6": ""}

        return resu

    async def secure(self, user, passwd):
        """ Nothing here """
        await asyncio.sleep(0)

    async def set_options(self, options={}):
        """ Nothing here """
        await asyncio.sleep(0)

    async def provision(self, ip, ssid, psk, ktype="none"):
        """Coroutine to perform provisioning

           As long as this method returns AGAIN, this coroutine must be called after switching
           wireless network if necessary. Each AGAIN shoud toggle between the device being provisioned
           cell and the original cell (nothig is done if the interface was not being used.

           If information is returnedits handling should be application specific.

            :param ip: IP address of the interface configured to access the device
            :type cmd: str
            :param ssid: The cell the device should connect to
            :type ssid: str
            :param psk: The key for the cell the device should connect to
            :type psk: str
            :param ktype: The key encyption type: wep, wpa, wpa2, none
            :type ktype: str
            :returns: a dictionary of information or AGAIN if needed
            :rtype: list
        """
        try:
            reader, writer = await asyncio.open_connection(
                ".".join(ip.split(".")[:-1] + ["1"]), 9999
            )
        except:
            self.go_on = False
            return {}
        message = (
            """{"netif":{"set_stainfo":{"ssid":"%s","password":"%s","key_type":%d}}}"""
        )
        message = message % (ssid, psk, ["none", "wep", "wpa", "wpa2"].index(ktype))
        logging.debug("TP-Link: Send: %r" % message)
        writer.write(encrypt(message))
        data = await reader.read(100)
        data = json.loads(decrypt(data[4:]))
        logging.debug("Received: %r" % data)
        writer.close()
        self.go_on = False
        resu = {}
        try:
            if data["netif"]["set_stainfo"]["err_code"] == 0:
                resu[data["netif"]["set_stainfo"]["mac"]] = {"type": "TP-Link"}
        except:
            pass
        finally:
            return resu


PluginObject = TPLink



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This is a library to connect to the AP provided by IoT devices and
# provision them.
#
# Copyright (c) 2018 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
#
#
#
BUGGYMODULE = ["8192cu"]


import asyncio as aio
import re, sys, os, importlib, logging
from collections import defaultdict

ifaceregexp = re.compile(r"^[0-9]+: (?P<iface>[a-zA-Z0-9]+): .*$")

ipregexps = [
    re.compile(r"^link/ether (?P<mac>[0-9a-fA-F:]+) brd .*$"),
    re.compile(r"^inet (?P<ip>[0-9.]+)/[0-9]+ .*$"),
    re.compile(r"^inet6 (?P<ipv6>[0-9:a-fA-F]+)/[0-9]+ .*$"),
]


def parse_ifaces(content):
    """Parsing the output of command "ip addr".

       This returns a dictionary of dictionaries.
        At the first level, the key is the interface name
            At the second level the values are lists and the keys are
                mac : List of mac addresses (usually 1)
                ip  : List if IPv4 addresses
                ipv6: List of IPv6 adresses

        :param content: A string to be parsed
        :type cmd: str
        :returns: Parsed content as a dictionary..
        :rtype: dict
    """
    loiface = {}
    addrs = defaultdict(list)
    iface = None
    lines = content.split("\n")
    for line in lines:
        if not line:
            continue
        niface = ifaceregexp.search(line)
        if niface:
            if iface and addrs:
                loiface[iface] = addrs
            addrs = defaultdict(list)
            iface = niface.groupdict()["iface"]
            continue
        line = line.strip()
        for expression in ipregexps:
            result = expression.search(line)
            if result is not None:
                for x, y in result.groupdict().items():
                    addrs[x].append(y)
                continue
    if iface and addrs:
        loiface[iface] = addrs
    logging.debug("--> {}".format(loiface))
    return loiface


class WiFiManager(object):
    def __init__(self):
        pass


class WPAWiFiManager(WiFiManager):
    def parse_cells(self, content):
        """This function parses the output of the command "/sbin/wpa_cli  scan_result"

        This function returns the parsed content as a list of dictionary.
        Each dictionary describes one scanned cell: ESSID,
        Encryption, BSSID


            :param content: A string to be parsed
            :type cmd: str
            :returns: Parsed content as a list..
            :rtype: list
        """
        foundheader = False
        lines = content.split("\n")
        locells = []
        for line in lines:
            if not line:
                continue
            thisline = [x for x in line.split("\t") if x.strip()]
            if "bssid" in line.lower() and "ssid" in line.lower():
                foundheader = True
                continue
            if foundheader:
                try:
                    acell = {}
                    acell["bssid"] = thisline[0]
                    if "[" in thisline[3]:
                        if "WPA2" in thisline[3]:
                            acell["encryption"] = "wpa2"
                        elif "WPA" in thisline[3]:
                            acell["encryption"] = "wpa"
                        elif "WEP" in thisline[3]:
                            acell["encryption"] = "wep"
                        else:
                            acell["encryption"] = "none"
                        acell["ssid"] = thisline[4]
                    else:
                        acell["encryption"] = "none"
                        acell["ssid"] = thisline[3]

                    locells.append(acell)
                except:
                    # In all likelyhood, a hidden SSID, ignore those
                    pass

        logging.debug("--> {}".format(locells))
        return locells

    async def gather_cellinfo(self, interfaces):

        cells = {}
        for iface in interfaces:
            iswifi = await run_cmd(["sudo", "/sbin/wpa_cli", "-i", iface, "scan"])
            if iswifi and iswifi.strip().lower() == "ok":
                await aio.sleep(4)
                cells[iface] = await run_cmd(
                    ["sudo", "/sbin/wpa_cli", "-i", iface, "scan_result"],
                    self.parse_cells,
                )
        return cells

    async def wifi_connect(self, iface, ssid, psk=None, is_wep=False):
        """Connect to the given wifi network.

            :param ssid: Name of the cell to connect to
            :type ssid: str
            :param psk: Key to use
            :type psk: str
            :returns: id of the network
            :rtype: int

        """
        netid = int(
            await run_cmd(["sudo", "/sbin/wpa_cli", "-i", iface, "add_network"])
        )
        if psk and not is_wep:
            wpacmd = [
                'set_network %d ssid "%s"' % (netid, ssid),
                'set_network %d psk "%s"' % (netid, psk),
                "enable_network %d" % netid,
                3,
                "quit",
            ]
        elif psk and is_wep:
            wpacmd = [
                'set_network %d ssid "%s"' % (netid, ssid),
                "set_network %d key_mgmt NONE" % (netid),
                "set_network %d wep_key0 %s" % (netid, psk),
                "set_network %d auth_alg SHARED" % (netid),
                "enable_network %d" % netid,
                3,
                "quit",
            ]
        else:
            wpacmd = [
                'set_network %d ssid "%s"' % (netid, ssid),
                "set_network %d key_mgmt NONE" % (netid),
                "enable_network %d" % netid,
                3,
                "quit",
            ]

        resul = await run_inter_cmd(["/sbin/wpa_cli", "-i", iface], wpacmd)
        logging.debug("{}".format(resul))
        return netid

    async def wifi_disconnect(self, iface, netid):
        """Connect to the given wifi network.

            :param netid: Network id, Got when connecting
            :type netid: int
            :returns: None
            :rtype: None

        """
        logging.debug("Disconnecting {}".format(iface))
        xx = await run_cmd(["sudo", "/sbin/wpa_cli", "-i", iface, "bss_flush"])
        xx = await run_cmd(
            ["sudo", "/sbin/wpa_cli", "-i", iface, "disable_network", str(netid)]
        )
        xx = await run_cmd(
            ["sudo", "/sbin/wpa_cli", "-i", iface, "remove_network", str(netid)]
        )
        xx = await run_cmd(["sudo", "/sbin/ip", "link", "set", "down", "dev", iface])
        xx = await run_cmd(["sudo", "/sbin/ip", "link", "set", "up", "dev", iface])
        # Work around bug with 8192cu
        for module in BUGGYMODULE:
            driver = await run_cmd(
                ["sudo", "/bin/readlink", "/sys/class/net/%s/device/driver" % iface]
            )
            driver = driver.strip()
            logging.debug(
                "Driver is {} it does{}end by {}".format(
                    driver, driver.endswith(module) and " " or " not ", module
                )
            )
            if driver.endswith(module):
                logging.debug("Work around for {}".format(module))
                xx = await run_cmd(["sudo", "/sbin/rmmod", module])
                xx = await aio.sleep(2)
                xx = await run_cmd(["sudo", "/sbin/modprobe", module])
                xx = await aio.sleep(1)

    async def wifi_reset(self, iface, ssid, psk):
        """Connect to the given wifi network.

            :returns: None
            :rtype: None

        """
        xx = await run_cmd(["sudo", "/sbin/wpa_cli", "-i", iface, "bss_flush"])
        xx = await run_cmd(["sudo", "/sbin/wpa_cli", "-i", iface, "reconfigure"])


class NMWiFiManager(WiFiManager):
    def parse_cells(self, content):
        """This function parses the output of the command "nmcli -t -f ssid,security,bssid -c no device wifi list"

        This function returns the parsed content as a list of dictionary.
        Each dictionary describes one scanned cell: ESSID,
        Encryption, BSSID


            :param content: A string to be parsed
            :type cmd: str
            :returns: Parsed content as a list..
            :rtype: list
        """
        locells = []
        lines = content.split("\n")
        for line in lines:
            if not line:
                continue
            thisline = [x for x in line.split(":")]
            if len(thisline) == 0 or thisline[0] in ["--", ""]:
                continue
            try:
                acell = {}
                acell["ssid"] = thisline[0]
                acell["bssid"] = ":".join(thisline[2:]).replace("\\", "").lower()
                if len(thisline) > 1:
                    if "WPA2" in thisline[1]:
                        acell["encryption"] = "wpa2"
                    elif "WPA" in thisline[1]:
                        acell["encryption"] = "wpa"
                    elif "WEP" in thisline[1]:
                        acell["encryption"] = "wep"
                    else:
                        acell["encryption"] = "none"
                else:
                    acell["encryption"] = "none"

                locells.append(acell)
            except:
                # In all likelyhood, a hidden SSID, ignore those
                pass

        logging.debug("--> {}".format(locells))
        return locells

    async def gather_cellinfo(self, interfaces):
        wifaces = []
        iswifi = await run_cmd(
            ["sudo", "/usr/bin/nmcli", "-t", "-f", "device,type", "device", "status"]
        )
        lines = iswifi.split("\n")
        for line in lines:
            if not line:
                continue
            thisline = [x for x in line.split(":")]
            if thisline[1] == "wifi":
                wifaces.append(thisline[0].strip())
        cells = {}
        iswifi = await run_cmd(["sudo", "/usr/bin/nmcli", "device", "wifi", "rescan"])
        # TDO check returned values
        await aio.sleep(4)
        allcells = await run_cmd(
            [
                "sudo",
                "/usr/bin/nmcli",
                "-t",
                "-f",
                "ssid,security,bssid",
                "-c",
                "no",
                "device",
                "wifi",
                "list",
            ],
            self.parse_cells,
        )
        for iface in wifaces:
            cells[iface] = allcells
        return cells

    async def wifi_connect(self, iface, ssid, psk=None, is_wep=False):
        """Connect to the given wifi network.

            :param ssid: Name of the cell to connect to
            :type ssid: str
            :param psk: Key to use
            :type psk: str
            :returns: id of the network
            :rtype: int

        """
        con = await run_cmd(
            [
                "sudo",
                "nmcli",
                "device",
                "wifi",
                "connect",
                ssid,
                "password",
                psk,
                "ifname",
                iface,
            ]
        )
        return None

    async def wifi_disconnect(self, iface, netid):
        """Connect to the given wifi network.

            :param netid: Network id, Got when connecting
            :type netid: int
            :returns: None
            :rtype: None

        """
        logging.debug("Disconnecting {}".format(iface))
        con = await run_cmd(["sudo", "nmcli", "device", "disconnect", iface])

    async def wifi_reset(self, iface, ssid, psk):
        """Connect to the given wifi network.

            :returns: None
            :rtype: None

        """
        await self.wifi_connect(iface, ssid, psk)


async def run_cmd(cmd, parse=None):
    """This coroutine runs a shell command within the asyncio framework

        :param cmd: A list with the various atoms of the command
        :type cmd: list
        :param parse: function to parse the command output.
        :type mac: func
        :returns: The output of the command possibly parsed.
        :rtype: object

    """
    proc = await aio.create_subprocess_exec(
        *cmd, stdout=aio.subprocess.PIPE, stderr=aio.subprocess.PIPE
    )
    stdout, stderr = await proc.communicate()
    if stderr:
        logging.info("Error whilst executing command:\n{}\n".format(stderr.decode()))
    if stdout:
        if parse:
            return parse(stdout.decode())
        else:
            return stdout.decode()


async def run_inter_cmd(cmd, icmd, parse=None):
    """This coroutine runs an interactive shell command within the asyncio framework

        :param cmd: A list with the various atoms of the command
        :type cmd: list
        :param icmd: A list of strings with the various commands to send down stdin. Integer indicate pauses in secs.
        :type icmd: list
        :param parse: function to parse the command output.
        :type mac: func
        :returns: The output of the command possibly parsed.
        :rtype: object

    """
    proc = await aio.create_subprocess_exec(
        *cmd,
        stdin=aio.subprocess.PIPE,
        stdout=aio.subprocess.PIPE,
        stderr=aio.subprocess.PIPE,
    )
    for scmd in icmd:
        if isinstance(scmd, int):
            await aio.sleep(scmd)
        else:
            proc.stdin.write(scmd.encode() + b"\n")
            await proc.stdin.drain()
    stdout, stderr = await proc.communicate()
    if stderr:
        logging.info("Error whilst executing command:\n{}\n".format(stderr.decode()))
    if stdout:
        if parse:
            return parse(stdout.decode())
        else:
            return stdout.decode()


async def check_system(neededcmd=["ip", "wpa_supplicant", "/sbin/wpa_cli"]):
    """Coroutine to check that a number of shell commands can be used.

        :param neededcmd: A list of the command to be checked
        :type neededcmd: list
        :returns: True or False, and a possible error message
        :rtype: 2-uple
    """
    for cmd in neededcmd:
        cpath = await run_cmd(["sudo", "which", cmd])
        if not cpath:
            logging.error("Error: Command {} is not available.".format(cmd))
            return False, "Error: Command {} is not available.".format(cmd)
    return True, ""


def load_plugins(needed=[]):
    """Load plugins

        :param needed: Name of the plugins to load.
        :type iface: list
        :returns: dictionary of plugin objects, key is the name
        :rtype: dict

    """
    dir_path = os.path.join(os.path.dirname(__file__), "plugins")
    # import parent module / namespace
    sys.path.append(os.path.dirname(__file__))
    importlib.import_module("plugins")
    logging.debug(f"Loading plugins from {needed}")
    loplugins = {}
    for filename in os.listdir(dir_path):
        name, ext = os.path.splitext(filename)
        if not needed or name in needed:
            if ext.endswith(".py"):
                try:
                    mod = importlib.import_module("." + name, package="plugins")
                    logging.debug("Loading plugin " + mod.PluginObject.name)
                    loplugins[mod.PluginObject.name] = mod.PluginObject
                except:
                    logging.error("Plugin {} could not be loaded.".format(name))

    return loplugins


class IoTProvision(object):
    def __init__(self, ssid, psk="", manager=NMWiFiManager()):
        """Create a IoTProvision object

        :param ssid: The SSID to connect the device to
        :type ssid: str
        :param psk: The shared secret to connect to SSID
        :type psk: str
        :returns: IoTProvision object.
        :rtype: IoTProvision
        """
        self.ssid = ssid
        self.ssid_encrypt = None
        self.psk = psk
        self.user = ""
        self.passw = ""
        self.cells = {}
        self.iface = None
        self.is_shared = False
        self.interfaces = {}
        self.wifimanager = manager

    def set_secure(self, user, passw):
        """Set the username and password to secure the device with.

        Individual plugins decide whether or not to use this

        :param user: The user
        :type user: str
        :param passw: The password
        :type passw: str
        :returns: Nothong.
        :rtype: None
        """
        self.user = user
        self.passw = passw

    async def gather_netinfo(self, iface=None, cells_too=True):
        """Coroutine to retrieve all relevant information about network interfaces and wifi cells available

            :param iface: Limit to the given interface
            :type iface: str
            :param cell_too: Whether or not to look for Wifi cells
            :type cell_too: bool
            :returns: The parsed output of "ip addr"
            :rtype: dict

        """
        if iface:
            self.interfaces = await run_cmd(
                ["sudo", "/sbin/ip", "addr", "show", "dev", iface], parse_ifaces
            )
        else:
            self.interfaces = await run_cmd(["sudo", "/sbin/ip", "addr"], parse_ifaces)

        if cells_too:
            self.cells = await self.wifimanager.gather_cellinfo(self.interfaces)
        return self.interfaces

    def select_iface(self, interfaces, cells):
        """Select a wirelless interface.

        This will try to find a wireless interface that is not configured. Failing
        that it will simply pick the first one.

            :param interfaces: A dictionary describing the interfaces
            :type interfaces: dict
            :param cells: A dictionary describing the wifi cells
            :type cells: dict
            :returns: An interface and whether or not this interface was configured
            :rtype: 2-uple

        """
        self.iface = None
        self.is_shared = False

        for iface in cells:
            if "ip" in interfaces[iface] or "ip6" in interfaces[iface]:
                continue
            # OK, This interface does not have an IP address
            self.iface = iface
            break

        if not self.iface:
            self.iface = [x for x in cells.keys()][0]
            # We will need to restore things at the end
            self.is_shared = True

        return self.iface != None

    async def wifi_connect(self, ssid, psk=None, is_wep=False):
        """Connect to the given wifi network.

            :param ssid: Name of the cell to connect to
            :type ssid: str
            :param psk: Key to use
            :type psk: str
            :returns: id of the network
            :rtype: int

        """
        return await self.wifimanager.wifi_connect(self.iface, ssid, psk, is_wep)

    async def wifi_disconnect(self, netid):
        """Connect to the given wifi network.

            :param netid: Network id, Got when connecting
            :type netid: int
            :returns: None
            :rtype: None

        """
        await self.wifimanager.wifi_disconnect(self.iface, netid)

    async def wifi_reset(self):
        """Connect to the given wifi network.

            :returns: None
            :rtype: None

        """
        await self.wifimanager.wifi_reset(self.iface, self.ssid, self.psk)

    async def provision(self, plugins=[], options={}):
        """Connect to the given wifi network.

            :param plugins: List of plugins to load. None means all.
            :type plugins: str
            :returns: List of returned info from the provisioning phase
            :rtype: list

        """
        result = {}
        loplugins = load_plugins(plugins)
        if self.ssid_encrypt == None:
            for cdef in self.cells[self.iface]:
                if cdef["ssid"] == self.ssid:
                    self.ssid_encrypt = cdef["encryption"]
                    break
        if self.ssid_encrypt == None:
            self.ssid_encrypt = "none"
        for aplug in loplugins.values():
            cando = aplug.can_handle([x["ssid"] for x in self.cells[self.iface]])
            #
            # logging.debug("==> {}".format(self.cells))
            for acell in cando:
                for cdef in self.cells[self.iface]:
                    if cdef["ssid"] == acell:
                        cencrypt = cdef["encryption"]
                        mac = cdef["bssid"]
                        break
                currcnt = 1
                myplug = aplug(mac)
                while myplug.go_on:
                    if currcnt == 1 or (currcnt % 2 and self.is_shared):
                        # Only connects if really needed
                        netid = await self.wifi_connect(
                            acell, cando[acell]["passwd"], cencrypt == "wep"
                        )
                    xx = await aio.sleep(5)
                    if cando[acell]["ip"] or cando[acell]["ipv6"]:
                        if cando[acell]["ip"]:
                            xx = await run_cmd(
                                [
                                    "sudo",
                                    "/sbin/ip",
                                    "addr",
                                    "add",
                                    cando[acell]["ip"],
                                    "dev",
                                    self.iface,
                                ]
                            )
                        if cando[acell]["ipv6"]:
                            xx = await run_cmd(
                                [
                                    "sudo",
                                    "/sbin/ip",
                                    "-6",
                                    "addr",
                                    "add",
                                    cando[acell]["ip"],
                                    "dev",
                                    self.iface,
                                ]
                            )
                        ifaceinfo = await self.gather_netinfo(self.iface, False)
                    else:
                        cnt = 5
                        while cnt:
                            cnt -= 1
                            ifaceinfo = await self.gather_netinfo(self.iface, False)
                            logging.debug(ifaceinfo)
                            if "ip" in ifaceinfo[self.iface]:
                                break
                            xx = await aio.sleep(1)

                        if "ip" not in ifaceinfo[self.iface]:
                            logging.error(
                                "Warning: Did not get an IP address from {}. Bailing out.".format(
                                    acell
                                )
                            )
                            xx = await self.wifi_disconnect(netid)
                            break

                    xx = await myplug.secure(self.user, self.passw)
                    if myplug.name in options:
                        xx = await myplug.set_options(options[myplug.name])
                    presu = await myplug.provision(
                        ifaceinfo[self.iface]["ip"][0],
                        self.ssid,
                        self.psk,
                        self.ssid_encrypt,
                    )
                    if not myplug.go_on or not self.is_shared:
                        xx = await self.wifi_disconnect(netid)
                    if not myplug.go_on:
                        result.update(presu)
                    currcnt += 1
        if self.is_shared:
            await self.wifi_reset()
        return result



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This plugin handles the provisioning of Sonoff devices running Tasmota
#
# Copyright (c) 2018 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
#
#
#
import asyncio
import aiohttp as aioh
import logging
import urllib

# DELAY=17
DELAY = 5


class Tasmota(object):

    name = "tasmota"

    def __init__(self, mac):
        """
        The go_on attribute must exist. It is set to False when provisioning is done

        :param mac: MAC address of the device being provisioned
        :type mac: str
        """
        self.go_on = True
        self.myid = "".join(mac.split(":")[-3:]).upper()
        self.myauth = None
        self.mac = mac
        self.options = {}

    @classmethod
    def can_handle(self, cells):
        """ Given a list of cell names, return a list of those it can handle

        :param cells: A list of cell names
        :type cmd: list
        :returns: A dictionary of those it can handle. The key is the cell's name, the value a dictionary
                  with a WIFI key "passwd" (empty is not needed) and IP addresses "ip" and "ipv6" (empty if use dhcp)
                  IP address must be of the form <address>/<bits>
        :rtype: dict
        """
        resu = {}
        for x in cells:
            if x.lower().strip().startswith("tasmota_"):
                resu[x] = {"passwd": "", "ip": "", "ipv6": ""}

        return resu

    async def secure(self, user, passwd):
        """ Setting the password... and remembering it for subsequent access
        With tasmota, we use the backlog command, to set everything at onece, so here we setup
        the options attribute
        """
        if passwd:
            self.myauth = aioh.BasicAuth(login=user, password=passwd)
            self.options["WebPassword"] = passwd
        await asyncio.sleep(0)

    async def set_options(self, options={}):
        """ Could set MQTT here

        :param options: A list of options to be set. Here MQTT setting is possible
        :type options: dict

        """
        logging.debug("options --> {}".format(options))
        # params = {}
        if "template" in options:
            self.options["Template"] = options["template"]
            self.options["Module"] = 0
            # params["t1"] = options["template"]
            # params["t2"] = 'on'
        elif "module" in options:
            self.options["Module"] = options["module"]

        if "mqtt" in options:
            if options["mqtt"] in [True, "on", 1]:
                self.options["SetOption3"] = 1
                for k, o in [
                    ("host", "MqttHost"),
                    ("port", "MqttPort"),
                    ("user", "MqttUser"),
                    ("password", "Mqttpassword"),
                ]:
                    if k in options:
                        self.options[o] = options[k]  # Set MQTT

                for k, o in [
                    ("client", "MqttClient"),
                    ("topic", "Topic"),
                    ("full topic", "FullTopic"),
                ]:
                    if k in options:
                        self.options[o] = options[k].replace(
                            "{mac}", "".join(self.mac.split(":")[-3:]).lower()
                        )  # Set MQTT
        await asyncio.sleep(0)

    async def provision(self, ip, ssid, psk, ktype="none"):
        """Coroutine to perform provisioning

           As long as this method returns AGAIN, this coroutine must be called after switching
           wireless network if necessary. Each AGAIN shoud toggle between the device being provisioned
           cell and the original cell (nothig is done if the interface was not being used.

           If information is returnedits handling should be application specific.

            :param ip: IP address of the interface configured to access the device
            :type cmd: str
            :param ssid: The cell the device should connect to
            :type ssid: str
            :param psk: The key for the cell the device should connect to
            :type psk: str
            :param ktype: The key encyption type: wep, wpa, wpa2, none
            :type ktype: str
            :returns: a dictionary of information or AGAIN if needed
            :rtype: list
        """
        resu = {}
        try:
            self.options.update(
                {
                    "SSID": ssid,
                    "PASSWORD1": psk,
                    "SSID2": ssid,
                    "PASSWORD2": psk,
                    "HOSTNAME": "1",
                }
            )
            param = {
                "cmnd": "Backlog "
                + ";".join([f"{x} {y}" for x, y in self.options.items()])
            }
            param = urllib.parse.urlencode(param, quote_via=urllib.parse.quote)
            doitagin = False
            async with aioh.ClientSession(auth=self.myauth) as session:
                async with session.request(
                    "get", "http://192.168.4.1/cm", params=param
                ) as resp:
                    logging.debug(resp.url)
                    logging.debug("Tasmota: Response status was {}".format(resp.status))
                    if resp.status == 401:
                        doitagin = True
            if doitagin:
                async with aioh.ClientSession(auth=None) as session:
                    async with session.request(
                        "get", "http://192.168.4.1/cm", params=param
                    ) as resp:
                        logging.debug(resp.url)
                        logging.debug(
                            "Tasmota: Response status was {}".format(resp.status)
                        )

            logging.debug("Tasmota: Set SSID and key")
            resu[self.mac] = {"type": "Tasmota"}
        except Exception as e:
            logging.debug(f"Tasmota: Could not set SSID: {e}")
            logging.exception(e)
        await asyncio.sleep(2)
        self.go_on = False
        return resu


PluginObject = Tasmota



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# This application looks for known AP (defined by plugins)
# and attempt to provision the respective IoT devices
#
# Copyright (c) 2018 François Wautier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
#
#
import argparse, os, sys, asyncio
import aioiotprov as aiop
import logging, json, shutil


class OptionString(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(OptionString, self).__init__(option_strings, dest, nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        options = {}

        try:
            for aval in values:
                key = aval.split(":")[0].strip()
                allvals = {}
                lovals = (":".join(aval.split("::")[1:])).split("||")
                for aval in lovals:
                    vkey = (aval.split("=="))[0].strip()
                    allvals[vkey] = (aval.split("=="))[1].strip()
                options[key] = allvals
        except:
            raise argparse.ArgumentTypeError(
                'Could not parse options "{}". You might want to add -- to end the options.'.format(
                    aval
                )
            )
        setattr(namespace, self.dest, options)


def main(args=None):
    # Needed arguments
    parser = argparse.ArgumentParser(description="Provision IoT devices")
    parser.add_argument("ssid", help="The SSID the device must be attaching to.")
    parser.add_argument(
        "passphrase", nargs="?", default="", help="The secret passphrase for the SSID."
    )
    parser.add_argument(
        "-u",
        "--user",
        default="",
        help="The user that will control access to the device",
    )
    parser.add_argument(
        "-p",
        "--password",
        default="",
        help="The password that will control access to the device.",
    )
    parser.add_argument(
        "-P", "--plugin", action="append", help="Select which plugin to use."
    )
    parser.add_argument(
        "-o",
        "--options",
        dest="options",
        default={},
        action=OptionString,
        nargs="*",
        metavar="<plugin>::opt1==val1||opt2==val2||..",
        help="Option to be passed to plugins.",
    )
    parser.add_argument(
        "-w",
        "--wpa",
        default=False,
        action="store_true",
        help="Use wpa_cli to control WiFi if command not found.",
    )
    parser.add_argument(
        "-n",
        "--nm",
        default=False,
        action="store_true",
        help="Use nmcli to control WiFi if command not found.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Print more information.",
    )
    parser.add_argument(
        "-j", "--json", default=False, action="store_true", help="Print result as json."
    )
    parser.add_argument(
        "-l",
        "--list",
        default=False,
        action="store_true",
        help="Just list the available SSID.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        default=False,
        action="store_true",
        help="Print debug information.",
    )

    try:
        opts = parser.parse_args()
    except Exception as e:
        parser.error("Error: " + str(e))
        sys.exit(1)

    if shutil.which("nmcli"):
        wifictl = aiop.NMWiFiManager()
    elif shutil.which("wpa_cli"):
        wifictl = aiop.WPAWiFiManager()
    elif opts.wpa:
        wifictl = aiop.WPAWiFiManager()
    elif opts.nm:
        wifictl = aiop.NMWiFiManager()
    else:
        print(
            "Error: Do not know how to managge WiFi interfaces. Neither 'nmcli' nor 'wpa_cli' are present."
        )
        sys.exit(1)

    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif opts.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.ERROR)

    loop = asyncio.get_event_loop()
    # First let's make sure we have the needed commands
    isok, errmsg = loop.run_until_complete(aiop.check_system())
    if not isok:
        print(errmsg)
        loop.close()
        sys.exit(1)

    provisioner = aiop.IoTProvision(opts.ssid, opts.passphrase, manager=wifictl)
    provisioner.set_secure(opts.user, opts.password)

    # First find out what interfaces are available
    interfaces = loop.run_until_complete(provisioner.gather_netinfo())

    if opts.list:
        loc = set()
        for aloc in provisioner.cells.values():
            for acell in [x["ssid"] for x in aloc]:
                loc.add(acell)
        print(json.dumps(list(loc)))
    else:
        if not provisioner.cells:
            print("Error: There seem to be no wifi adapter")
            sys.exit(2)

        isset = provisioner.select_iface(interfaces, provisioner.cells)
        if not isset:
            print("Error: No interface could be selected")

        else:
            logging.info(
                "Using interface {} with restore {}".format(
                    provisioner.iface, provisioner.is_shared
                )
            )
            resu = loop.run_until_complete(
                provisioner.provision(options=opts.options, plugins=opts.plugin)
            )
            logging.info("Got: {}".format(resu))
            if opts.json:
                print(json.dumps(resu))
    loop.close()


if __name__ == "__main__":
    main()
